body {
    font: @baseFont;
}

p {
    white-space: pre-wrap;
}

.action {
    color: @action;
    font-style: italic;
    font-weight: bold;
}

.typing {
    color: #8d8d8d;
}

.quote {
    color: #279419;
}

.alert {
    margin-left: 0px;
    margin-right: 0px;
    background-color: @orange;
}

.alert_name {
    background-color: @orange;
    font: @bigBold;
}

a {
    color: @link;
    font-weight: bold
}
