/*
    Copyright © 2019 by The qTox Project Contributors

    This file is part of qTox, a Qt-based graphical interface for Tox.

    qTox is libre software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    qTox is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with qTox.  If not, see <http://www.gnu.org/licenses/>.
*/

#ifndef SETTINGS_H
#define SETTINGS_H

#include <QString>

#ifdef Q_OS_WIN
#include <windows.h>
#endif

class Settings
{
public:
    Settings();
    ~Settings();

    QString getSettingsDirPath() const; ///< The returned path ends with a directory separator
#ifdef Q_OS_WIN
    HANDLE getPrimaryToken() const; ///< Used to impersonnate the unelevated user
#endif

private:
    bool portable;
    static constexpr const char* SETTINGS_FILE = "qtox.ini";
#ifdef Q_OS_WIN
    HANDLE hPrimaryToken;
#endif
};

#endif // SETTINGS_H
