QToolTip
{
    /* explicit border width is required https://bugreports.qt.io/browse/QTBUG-41313 */
    border: 0px;
    color: black;
    background: #ffffdc;
}

QDialog
{
    background: white;
}

QWidget#contentWidget
{
    background: white;
}

QTabWidget
{
    background-color: white;
}

QTabBar::tab:selected 
{
    background: #2d3136;
    color: #f1f1f1;
}

QTabBar::tab:!selected 
{
    background: #d0d1d1;
    color: #5e5e5e;
}

QLabel
{
    color: @mainText;
}

QListView
{
    color: @mainText;
}

QTextEdit, QPlainTextEdit
{
    border-color: @groundExtra;
    border-style: solid;
    border-width: 1px 0 1px 1px;
    background: white;
    color: @mainText;
}

QListWidget
{
    background-color: white;
}

QMessageBox
{
    background-color: white;
}

QCheckBox
{
    color: black;
}

QCheckBox:disabled
{
    color: grey;
}

QSpinBox, QDoubleSpinBox
{
    color: @mainText;
    background-color: white;
}

QSpinBox:disabled, QDoubleSpinBox:disabled
{
    color: @mainText;
    background-color: lightGrey;
}

QGroupBox
{
    color: black;
    background-color: white;
    font: @bigBold;
}

QComboBox
{
    color: black;
    background-color: white;
}

QComboBox QAbstractItemView {
    background-color: white;
}

QLineEdit
{
    color: @mainText;
    background-color: white;
}

QLineEdit:disabled
{
    color: @mainText;
    background-color: lightGrey;
}

QScrollArea
{
    background-color: white;
}

QScrollArea > QWidget > QWidget
{
    background: transparent;
}

QScrollArea::corner
{
    background-color: white;
    border: none;
}

QScrollBar:vertical
{
    background: transparent;
    width: 12px;
    margin-top: 2px;
    margin-bottom: 2px;
}

/* using last is a bit of a hack, but QTabBar otherwise doesn't allow selecting single tabs */
QTabBar::tab:last:!selected[update-available=true]
{
    background-color: #80c580;
}

QPushButton#updateAvailableButton
{
    background-color: #21da21;
}

QScrollBar::handle:vertical
{
    background-color: #d1d1d1;
    min-height: 20px;
    border-radius: 3px;
    margin-left: 2px;
}

QScrollBar::handle:vertical:hover
{
    background-color: #e3e3e3;
}

QScrollBar::handle:vertical:pressed
{
    background-color: #b1b1b1;
}

QScrollBar::add-line:vertical
{
    background-color: white;
    height: 0px;
    subcontrol-position: bottom;
    subcontrol-origin: margin;
}

QScrollBar::sub-line:vertical
{
    background-color: white;
    height: 0px;
    subcontrol-position: top;
    subcontrol-origin: margin;
}

QScrollBar:QScrollBar::down-arrow:vertical
{
    width: 10;
    height: 10px;
    background-color: white;
}

QScrollBar:QScrollBar::up-arrow:vertical
{
    width: 10px;
    height: 10px;
    background-color: white;
}

QScrollBar::add-page:vertical, QScrollBar::sub-page:vertical
{
    background: none;
}

QScrollBar:horizontal
{
    background-color: white;
    height: 10px;
    margin: 0 2px 0 2px;
}

QScrollBar::handle:horizontal
{
    background-color: #d1d1d1;
    min-width: 20px;
    border-radius: 2px;
}

QScrollBar::handle:horizontal:hover
{
    background-color: #e3e3e3;
}

QScrollBar::handle:horizontal:pressed
{
    background-color: #b1b1b1;
}

QScrollBar::add-line:horizontal
{
    background-color: white;
    width: 0px;
    subcontrol-position: right;
    subcontrol-origin: margin;
}

QScrollBar::sub-line:horizontal
{
    background-color: white;
    width: 0px;
    subcontrol-position: left;
    subcontrol-origin: margin;
}

QScrollBar:QScrollBar::down-arrow:horizontal
{
    width: 10;
    height: 10px;
    background-color: white;
}

QScrollBar:QScrollBar::up-arrow:horizontal
{
    width: 10px;
    height: 10px;
    background-color: white;
}

QScrollBar::add-page:horizontal, QScrollBar::sub-page:horizontal
{
    background: none;
}

QRadioButton
{
    background: white;
    color: black;
}

QPushButton
{
    background: #ebebeb;
    color: @mainText;
}

QSplitter:handle
{
    color: white;
    background-color: white;
}
