QPushButton
{
    background-color: #6bc260;
    background-image: url("@getImagePath(screenshotButton/screenshotButton.svg)");
    background-repeat: none;
    background-position: center;
    border-top-left-radius: 5px;
    border: none;
    width: 24px;
    height: 24px;
}

QPushButton:hover
{
    background-color: #79c76f;
}

QPushButton:pressed
{
    background-color: #51b244;
}

QPushButton[enabled="false"]
{
    background-color: #919191;
}

QPushButton:focus
{
    outline: none;
}
