QLineEdit
{
    background: none;
    background-color: @themeMedium;
    color: white;
    border: 0px;
    border-radius: 4px;
}

QToolButton
{
    background: none;
    background-color: @themeMedium;
    color: white;
    border-style: none;
    border-radius: 4px;
}

QToolButton:pressed
{
    background-color: @themeMediumDark;
    border-radius: 4px;
    color: white;
}

QToolButton::menu-indicator
{
    image: none
}

QPushButton#green
{
    background: none;
    background-color: #6bc260;
    color: white;
    border-style: none;
    border-radius: 4px;
    padding: 4px;
    margin: 4px 8px;
}

QPushButton#green:hover
{
    background-color: #79c76f;
}

QPushButton#green:pressed
{
    background-color: #51b244;
}

/**
  Uncomment this after https://github.com/qTox/qTox/pull/1640
  is merged!
QComboBox:down-arrow
{
    image: url(ui/css/down_arrow.png);
}
**/

QListView
{
    background-color: @themeLight;
    border-style: none;
    border-radius: 4px;
}

#statusPanel
{
  background-color: @themeDark;
}

#statusPanel > #statusHead
{
  background-color: @themeDark;
}

#statusPanel > #statusHead > #nameLabel
{
  background-color: @themeDark;
  font: @extraBig;
  color: @groundBase;
}

#statusPanel > #statusHead > #statusLabel
{
  background-color: @themeDark;
  font: @medium;
  color: @groundExtra;
}

#statusPanel > #statusHead > #statusButton
{
    background: none;
    background-color: @themeMedium;
    border: none;
    border-radius: 6px;
    width: 20px;
    height: 40px;
}

/*Bugged in Qt, but it's probably better to leave enabled so that users can tell it's clickable*/
#statusPanel > #statusHead > #statusButton:hover
{
  background-color: @themeLight;
}

#statusPanel > #statusHead > #statusButton:pressed
{
  background-color: @themeMedium;
}

#statusPanel > #statusHead > #statusButton:focus
{
  outline: none;
}

#statusPanel > #statusHead > #statusButton::menu-indicator {image: none;}

#statusPanel > #statusHead > #statusButton::menu-indicator:pressed, #statusPanel > #statusHead > #statusButton::menu-indicator:open
{
  image: url("@getImagePath(statusButton/menu_indicator.svg)");
  subcontrol-origin: padding;
  subcontrol-position: bottom center;
  position: relative;
  bottom: 2px;
}
